@@ -10,6 +10,14 @@ class Mode
     @@keymaps[self] = keymap
   end
 
+  def self.keymap
+    @@keymaps[self] || register_keymap
+  end
+
+  def self.keymaps
+    @@keymaps
+  end
+
   def initialize
     @buffer = nil
   end
@@ -38,7 +46,7 @@ class Mode
   end
 
   def resolve_input c
-    ancestors.each do |klass| # try all keymaps in order of ancestry
+    self.class.ancestors.each do |klass| # try all keymaps in order of ancestry
       next unless @@keymaps.member?(klass)
       action = BufferManager.resolve_input_with_keymap c, @@keymaps[klass]
       return action if action
@@ -54,7 +62,7 @@ class Mode
 
   def help_text
     used_keys = {}
-    ancestors.map do |klass|
+    self.class.ancestors.map do |klass|
       km = @@keymaps[klass] or next
       title = "Keybindings from #{Mode.make_name klass.name}"
       s = <<EOS
@@ -75,7 +83,8 @@ EOS
 ### helper functions
 
   def save_to_file fn, talk=true
-    if File.exists? fn
+    FileUtils.mkdir_p File.dirname(fn)
+    if File.exist? fn
       unless BufferManager.ask_yes_or_no "File \"#{fn}\" exists. Overwrite?"
         info "Not overwriting #{fn}"
         return
@@ -94,37 +103,42 @@ EOS
   end
 
   def pipe_to_process command
-    Open3.popen3(command) do |input, output, error|
-      err, data, * = IO.select [error], [input], nil
+    begin
+      Open3.popen3(command) do |input, output, error|
+        err, data, * = IO.select [error], [input], nil
 
-      unless err.empty?
-        message = err.first.read
-        if message =~ /^\s*$/
-          warn "error running #{command} (but no error message)"
-          BufferManager.flash "Error running #{command}!"
-        else
-          warn "error running #{command}: #{message}"
-          BufferManager.flash "Error: #{message}"
+        unless err.empty?
+          message = err.first.read
+          if message =~ /^\s*$/
+            warn "error running #{command} (but no error message)"
+            BufferManager.flash "Error running #{command}!"
+          else
+            warn "error running #{command}: #{message}"
+            BufferManager.flash "Error: #{message}"
+          end
+          return nil, false
         end
-        return
-      end
 
-      data = data.first
-      data.sync = false # buffer input
+        data = data.first
+        data.sync = false # buffer input
 
-      yield data
-      data.close # output will block unless input is closed
+        yield data
+        data.close # output will block unless input is closed
 
-      ## BUG?: shows errors or output but not both....
-      data, * = IO.select [output, error], nil, nil
-      data = data.first
+        ## BUG?: shows errors or output but not both....
+        data, * = IO.select [output, error], nil, nil
+        data = data.first
 
-      if data.eof
-        BufferManager.flash "'#{command}' done!"
-        nil
-      else
-        data.read
+        if data.eof
+          BufferManager.flash "'#{command}' done!"
+          return nil, true
+        else
+          return data.read, true
+        end
       end
+    rescue Errno::ENOENT
+      # If the command is invalid
+      return nil, false
     end
   end
 end
