@@ -1,3 +1,5 @@
+# encoding: utf-8
+
 require 'thread'
 require 'lockfile'
 require 'mime/types'
@@ -5,15 +7,9 @@ require 'pathname'
 require 'set'
 require 'enumerator'
 require 'benchmark'
-
-## time for some monkeypatching!
-class Symbol
-  unless method_defined? :to_proc
-    def to_proc
-      proc {  |obj, *args| obj.send(self, *args) }
-    end
-  end
-end
+require 'unicode'
+require 'unicode/display_width'
+require 'fileutils'
 
 class Lockfile
   def gen_lock_id
@@ -30,7 +26,7 @@ class Lockfile
   def dump_lock_id lock_id = @lock_id
       "host: %s\npid: %s\nppid: %s\ntime: %s\nuser: %s\npname: %s\n" %
         lock_id.values_at('host','pid','ppid','time','user', 'pname')
-    end
+  end
 
   def lockinfo_on_disk
     h = load_lock_id IO.read(path)
@@ -42,6 +38,17 @@ class Lockfile
   def touch_yourself; touch path end
 end
 
+class File
+  # platform safe file.link which attempts a copy if hard-linking fails
+  def self.safe_link src, dest
+    begin
+      File.link src, dest
+    rescue
+      FileUtils.copy src, dest
+    end
+  end
+end
+
 class Pathname
   def human_size
     s =
@@ -74,7 +81,7 @@ module RMail
     end
 
     def charset
-      if header.field?("content-type") && header.fetch("content-type") =~ /charset="?(.*?)"?(;|$)/i
+      if header.field?("content-type") && header.fetch("content-type") =~ /charset\s*=\s*"?(.*?)"?(;|$)/i
         $1
       end
     end
@@ -111,12 +118,47 @@ module RMail
       # end
     end
   end
-end
 
-class Range
-  ## only valid for integer ranges (unless I guess it's exclusive)
-  def size
-    last - first + (exclude_end? ? 0 : 1)
+  class Header
+
+    # Convert to ASCII before trying to match with regexp
+    class Field
+
+      class << self
+        def parse(field)
+          field = field.dup.to_s
+          field = field.fix_encoding!.ascii
+          if field =~ EXTRACT_FIELD_NAME_RE
+            [ $1, $'.chomp ]
+          else
+            [ "", Field.value_strip(field) ]
+          end
+        end
+      end
+    end
+
+    ## Be more cautious about invalid content-type headers
+    ## the original RMail code calls
+    ## value.strip.split(/\s*;\s*/)[0].downcase
+    ## without checking if split returned an element
+
+    # This returns the full content type of this message converted to
+    # lower case.
+    #
+    # If there is no content type header, returns the passed block is
+    # executed and its return value is returned.  If no block is passed,
+    # the value of the +default+ argument is returned.
+    def content_type(default = nil)
+      if value = self['content-type'] and ct = value.strip.split(/\s*;\s*/)[0]
+        return ct.downcase
+      else
+        if block_given?
+          yield
+        else
+          default
+        end
+      end
+    end
   end
 end
 
@@ -141,17 +183,6 @@ class Module
 end
 
 class Object
-  def ancestors
-    ret = []
-    klass = self.class
-
-    until klass == Object
-      ret << klass
-      klass = klass.superclass
-    end
-    ret
-  end
-
   ## "k combinator"
   def returning x; yield x; x; end
 
@@ -208,13 +239,15 @@ class Object
 end
 
 class String
-  ## nasty multibyte hack for ruby 1.8. if it's utf-8, split into chars using
-  ## the utf8 regex and count those. otherwise, use the byte length.
   def display_length
-    if RUBY_VERSION < '1.9.1' && ($encoding == "UTF-8" || $encoding == "utf8")
-      scan(/./u).size
-    else
-      size
+    @display_length ||= Unicode::DisplayWidth.of(self)
+  end
+
+  def slice_by_display_length len
+    each_char.each_with_object "" do |c, buffer|
+      len -= Unicode::DisplayWidth.of(c)
+      return buffer if len < 0
+      buffer << c
     end
   end
 
@@ -301,20 +334,66 @@ class String
   def wrap len
     ret = []
     s = self
-    while s.length > len
-      cut = s[0 ... len].rindex(/\s/)
+    while s.display_length > len
+      slice = s.slice_by_display_length(len)
+      cut = slice.rindex(/\s/)
       if cut
         ret << s[0 ... cut]
         s = s[(cut + 1) .. -1]
       else
-        ret << s[0 ... len]
-        s = s[len .. -1]
+        ret << slice
+        s = s[slice.length .. -1]
       end
     end
     ret << s
   end
 
+  # Fix the damn string! make sure it is valid utf-8, then convert to
+  # user encoding.
+  def fix_encoding!
+    # first try to encode to utf-8 from whatever current encoding
+    encode!('UTF-8', :invalid => :replace, :undef => :replace)
+
+    # do this anyway in case string is set to be UTF-8, encoding to
+    # something else (UTF-16 which can fully represent UTF-8) and back
+    # ensures invalid chars are replaced.
+    encode!('UTF-16', 'UTF-8', :invalid => :replace, :undef => :replace)
+    encode!('UTF-8', 'UTF-16', :invalid => :replace, :undef => :replace)
+
+    fail "Could not create valid UTF-8 string out of: '#{self.to_s}'." unless valid_encoding?
+
+    # now convert to $encoding
+    encode!($encoding, :invalid => :replace, :undef => :replace)
+
+    fail "Could not create valid #{$encoding.inspect} string out of: '#{self.to_s}'." unless valid_encoding?
+
+    self
+  end
+
+  # transcode the string if original encoding is know
+  # fix if broken.
+  def transcode to_encoding, from_encoding
+    begin
+      encode!(to_encoding, from_encoding, :invalid => :replace, :undef => :replace)
+
+      unless valid_encoding?
+        # fix encoding (through UTF-8)
+        encode!('UTF-16', from_encoding, :invalid => :replace, :undef => :replace)
+        encode!(to_encoding, 'UTF-16', :invalid => :replace, :undef => :replace)
+      end
+
+    rescue Encoding::ConverterNotFoundError
+      debug "Encoding converter not found for #{from_encoding.inspect} or #{to_encoding.inspect}, fixing string: '#{self.to_s}', but expect weird characters."
+      fix_encoding!
+    end
+
+    fail "Could not create valid #{to_encoding.inspect} string out of: '#{self.to_s}'." unless valid_encoding?
+
+    self
+  end
+
   def normalize_whitespace
+    fix_encoding!
     gsub(/\t/, "    ").gsub(/\r/, "")
   end
 
@@ -356,19 +435,8 @@ class String
         out << b.chr
       end
     end
-    out.force_encoding Encoding::UTF_8 if out.respond_to? :force_encoding
-    out
-  end
-
-  def transcode src_encoding=$encoding
-    Iconv.easy_decode $encoding, src_encoding, self
-  end
-
-  unless method_defined? :ascii_only?
-    def ascii_only?
-      size.times { |i| return false if self[i] & 128 != 0 }
-      return true
-    end
+    out = out.fix_encoding! # this should now be an utf-8 string of ascii
+                           # compat chars.
   end
 end
 
@@ -387,18 +455,18 @@ class Numeric
 
   def to_human_size
     if self < 1024
-      to_s + "b"
+      to_s + "B"
     elsif self < (1024 * 1024)
-      (self / 1024).to_s + "k"
+      (self / 1024).to_s + "KiB"
     elsif self < (1024 * 1024 * 1024)
-      (self / 1024 / 1024).to_s + "m"
+      (self / 1024 / 1024).to_s + "MiB"
     else
-      (self / 1024 / 1024 / 1024).to_s + "g"
+      (self / 1024 / 1024 / 1024).to_s + "GiB"
     end
   end
 end
 
-class Fixnum
+class Integer
   def to_character
     if self < 128 && self >= 0
       chr
@@ -407,12 +475,6 @@ class Fixnum
     end
   end
 
-  unless method_defined?(:ord)
-    def ord
-      self
-    end
-  end
-
   ## hacking the english language
   def pluralize s
     to_s + " " +
@@ -446,7 +508,7 @@ module Enumerable
   end
 
   def sum; inject(0) { |x, y| x + y }; end
-  
+
   def map_to_hash
     ret = {}
     each { |x| ret[x] = yield(x) }
@@ -496,10 +558,6 @@ module Enumerable
   end
 end
 
-unless Object.const_defined? :Enumerator
-  Enumerator = Enumerable::Enumerator
-end
-
 class Array
   def flatten_one_level
     inject([]) { |a, e| a + e }
@@ -514,77 +572,6 @@ class Array
   def nonempty?; !empty? end
 end
 
-class Time
-  def to_indexable_s
-    sprintf "%012d", self
-  end
-
-  def nearest_hour
-    if min < 30
-      self
-    else
-      self + (60 - min) * 60
-    end
-  end
-
-  def midnight # within a second
-    self - (hour * 60 * 60) - (min * 60) - sec
-  end
-
-  def is_the_same_day? other
-    (midnight - other.midnight).abs < 1
-  end
-
-  def is_the_day_before? other
-    other.midnight - midnight <=  24 * 60 * 60 + 1
-  end
-
-  def to_nice_distance_s from=Time.now
-    later_than = (self < from)
-    diff = (self.to_i - from.to_i).abs.to_f
-    text = 
-      [ ["second", 60],
-        ["minute", 60],
-        ["hour", 24],
-        ["day", 7],
-        ["week", 4.345], # heh heh
-        ["month", 12],
-        ["year", nil],
-      ].argfind do |unit, size|
-        if diff.round <= 1
-          "one #{unit}"
-        elsif size.nil? || diff.round < size
-          "#{diff.round} #{unit}s"
-        else
-          diff /= size.to_f
-          false
-        end
-      end
-    if later_than
-      text + " ago"
-    else
-      "in " + text
-    end  
-  end
-
-  TO_NICE_S_MAX_LEN = 9 # e.g. "Yest.10am"
-  def to_nice_s from=Time.now
-    if year != from.year
-      strftime "%b %Y"
-    elsif month != from.month
-      strftime "%b %e"
-    else
-      if is_the_same_day? from
-        strftime("%l:%M%p").downcase # emulate %P (missing on ruby 1.8 darwin)
-      elsif is_the_day_before? from
-        "Yest."  + nearest_hour.strftime("%l%p").downcase # emulate %P
-      else
-        strftime "%b %e"
-      end
-    end
-  end
-end
-
 ## simple singleton module. far less complete and insane than the ruby standard
 ## library one, but it automatically forwards methods calls and allows for
 ## constructors that take arguments.
@@ -592,31 +579,41 @@ end
 ## classes that inherit this can define initialize. however, you cannot call
 ## .new on the class. To get the instance of the class, call .instance;
 ## to create the instance, call init.
-module Singleton
-  module ClassMethods
-    def instance; @instance; end
-    def instantiated?; defined?(@instance) && !@instance.nil?; end
-    def deinstantiate!; @instance = nil; end
-    def method_missing meth, *a, &b
-      raise "no #{name} instance defined in method call to #{meth}!" unless defined? @instance
+module Redwood
+  module Singleton
+    module ClassMethods
+      def instance; @instance; end
+      def instantiated?; defined?(@instance) && !@instance.nil?; end
+      def deinstantiate!; @instance = nil; end
+      def method_missing meth, *a, &b
+        raise "no #{name} instance defined in method call to #{meth}!" unless defined? @instance
 
-      ## if we've been deinstantiated, just drop all calls. this is
-      ## useful because threads that might be active during the
-      ## cleanup process (e.g. polling) would otherwise have to
-      ## special-case every call to a Singleton object
-      return nil if @instance.nil?
+        ## if we've been deinstantiated, just drop all calls. this is
+        ## useful because threads that might be active during the
+        ## cleanup process (e.g. polling) would otherwise have to
+        ## special-case every call to a Singleton object
+        return nil if @instance.nil?
 
-      @instance.send meth, *a, &b
+        # Speed up further calls by defining a shortcut around method_missing
+        if meth.to_s[-1,1] == '='
+          # Argh! Inconsistency! Setters do not work like all the other methods.
+          class_eval "def self.#{meth}(a); @instance.send :#{meth}, a; end"
+        else
+          class_eval "def self.#{meth}(*a, &b); @instance.send :#{meth}, *a, &b; end"
+        end
+
+        @instance.send meth, *a, &b
+      end
+      def init *args
+        raise "there can be only one! (instance)" if instantiated?
+        @instance = new(*args)
+      end
     end
-    def init *args
-      raise "there can be only one! (instance)" if instantiated?
-      @instance = new(*args)
-    end
-  end
 
-  def self.included klass
-    klass.private_class_method :allocate, :new
-    klass.extend ClassMethods
+    def self.included klass
+      klass.private_class_method :allocate, :new
+      klass.extend ClassMethods
+    end
   end
 end
 
@@ -629,7 +626,7 @@ end
 ##   attr_accessor :val
 ##   def initialize; @val = 0 end
 ## end
-## 
+##
 ## h = Hash.new { C.new }
 ## h[:a].val # => 0
 ## h[:a].val = 1
@@ -656,32 +653,6 @@ class SavingHash
   defer_all_other_method_calls_to :hash
 end
 
-class OrderedHash < Hash
-  alias_method :store, :[]=
-  alias_method :each_pair, :each
-  attr_reader :keys
-
-  def initialize *a
-    @keys = []
-    a.each { |k, v| self[k] = v }
-  end
-
-  def []= key, val
-    @keys << key unless member?(key)
-    super
-  end
-
-  def values; keys.map { |k| self[k] } end
-  def index key; @keys.index key end
-
-  def delete key
-    @keys.delete key
-    super
-  end
-
-  def each; @keys.each { |k| yield k, self[k] } end
-end
-
 ## easy thread-safe class for determining who's the "winner" in a race (i.e.
 ## first person to hit the finish line
 class FinishLine
@@ -695,27 +666,3 @@ class FinishLine
   end
 end
 
-class Iconv
-  def self.easy_decode target, orig_charset, text
-    if text.respond_to? :force_encoding
-      text = text.dup
-      text.force_encoding Encoding::BINARY
-    end
-    charset = case orig_charset
-      when /UTF[-_ ]?8/i then "utf-8"
-      when /(iso[-_ ])?latin[-_ ]?1$/i then "ISO-8859-1"
-      when /iso[-_ ]?8859[-_ ]?15/i then 'ISO-8859-15'
-      when /unicode[-_ ]1[-_ ]1[-_ ]utf[-_]7/i then "utf-7"
-      when /^euc$/i then 'EUC-JP' # XXX try them all?
-      when /^(x-unknown|unknown[-_ ]?8bit|ascii[-_ ]?7[-_ ]?bit)$/i then 'ASCII'
-      else orig_charset
-    end
-
-    begin
-      returning(Iconv.iconv(target + "//IGNORE", charset, text + " ").join[0 .. -2]) { |str| str.check }
-    rescue Errno::EINVAL, Iconv::InvalidEncoding, Iconv::InvalidCharacter, Iconv::IllegalSequence, String::CheckError
-      debug "couldn't transcode text from #{orig_charset} (#{charset}) to #{target} (#{text[0 ... 20].inspect}...): got #{$!.class} (#{$!.message})"
-      text.ascii
-    end
-  end
-end
